/*
 * Jindolf main class
 *
 * Copyright(c) 2008 olyutorskii
 * $Id: Jindolf.java 88 2008-07-10 17:24:09Z olyutorskii $
 */

package jp.sourceforge.jindolf;

import java.awt.Dimension;
import java.awt.GraphicsEnvironment;
import java.awt.Toolkit;
import java.awt.Window;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JWindow;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

/**
 * Jindolf スタートアップクラス。
 * 
 * コンストラクタは無いよ。
 * アプリ開始はstaticメソッド{@link #main(String[])}呼び出しから。
 */
public class Jindolf{
    
    public static final String version = "2.5.2";

    public static final String license = "The MIT License";
    public static final String contact = "http://jindolf.sourceforge.jp/";
    
    public static final Date initDate;
    public static final long initNano;
    
    public static final ClassLoader loader;
    public static final Class selfClass;
    public static final Package selfPackage;
    public static final Package jrePackage;
    
    public static final Runtime runtime;
    public static SecurityManager secManager;
    public static final Logger logger;
    
    public static final String title;
    public static final String author;
    public static final String copyright;

    private static Dimension initGeometry = new Dimension(800, 600);
    
    static{
        initDate = new Date();
        initNano = System.nanoTime();

        selfClass = Jindolf.class;
        selfPackage = selfClass.getPackage();
        jrePackage = java.lang.Object.class.getPackage();

        runtime = Runtime.getRuntime();
        secManager = System.getSecurityManager();
        logger = Logger.getLogger(selfPackage.getName());

        ClassLoader thisLoader;
        try{
            thisLoader = selfClass.getClassLoader();
        }catch(SecurityException e){
            logger.log(Level.WARNING, "クラスローダの取得に失敗しました", e);
            thisLoader = null;
        }
        loader = thisLoader;

        String implTitle = selfPackage.getImplementationTitle();
        if(implTitle != null){
            title = implTitle;
        }else{
            title = "Jindolf";
        }

        String implVersion = selfPackage.getImplementationVersion();
        if(implVersion != null){
            if( ! version.equals(implVersion) ){
                logger.severe("バージョン定義上のコンパイルミスが検出されました");
                throw new ExceptionInInitializerError(
                        "unmatch package version");
            }
        }

        String implVendor = selfPackage.getImplementationVendor();
        if(implVendor != null){
            author = implVendor;
        }else{
            author = "olyutorskii";
        }

        copyright = "Copyright(c) 2008 " + author;
        
        logger.info(title +"/"+ version +" "+ copyright +" ("+ license +")"
                + " が " + initDate + " にロードされました。 " );
    }

    /**
     * クラスローダを介してリソースからの入力を生成する。
     * @param name リソース名
     * @return リソースからの入力
     */
    public static InputStream getResourceAsStream(String name){
        return selfClass.getResourceAsStream(name);
    }
    
    /**
     * クラスローダを介してリソース読み込み用URLを生成する。
     * @param name リソース名
     * @return URL
     */
    public static URL getResource(String name){
        return selfClass.getResource(name);
    }
    
    /**
     * 任意のクラス群に対して一括ロード／初期化を単一スレッドで順に行う。
     * どーしてもクラス初期化の順序に依存する障害が発生する場合や
     * クラス初期化のオーバーヘッドでGUIの操作性が損なわれるときなどにどうぞ。
     * 
     * @throws java.lang.LinkageError
     * @throws java.lang.ExceptionInInitializerError
     */
    private static void preLoadClass()
            throws LinkageError, ExceptionInInitializerError{
        String[] classes = {
            "java.lang.Object",
            "jp.sourceforge.jindolf.JdfEditorKit",
            "jp.sourceforge.jindolf.TabBrowser",
            "jp.sourceforge.jindolf.TopFrameView",
            "java.net.HttpURLConnection",
            "java.text.SimpleDateFormat",
            "javax.swing.JEditorPane",
            "javax.swing.text.html.CSS",
            "javax.swing.text.html.StyleSheet",
            "javax.swing.text.html.HTMLDocument",
            "javax.swing.text.html.BlockView",
            "javax.swing.text.html.ParagraphView",
            "javax.swing.text.html.InlineView",
            "javax.swing.text.html.parser.DocumentParser",
        };

        for(String className : classes){
            try{
                if(loader != null){
                    Class.forName(className, true, loader);
                }else{
                    Class.forName(className);
                }
            }catch(ClassNotFoundException e){
                logger.log(Level.WARNING,
                           "クラスの明示的ロードに失敗しました",
                           e);
            }
        }

        return;
    }

    /**
     * JRE1.5相当のランタイムライブラリ環境が提供されているか判定する。
     * 
     * @return JRE1.5相当ならtrueを返す。
     */
    public static boolean hasRuntime5(){
        boolean result;
        try{
            result = jrePackage.isCompatibleWith("1.5");
        }catch(NumberFormatException e){
            assert false;
            return false;
        }
        return result;
    }

    /**
     * JRE1.6相当のランタイムライブラリ環境が提供されているか判定する。
     * 
     * @return JRE1.6相当ならtrueを返す。
     */
    public static boolean hasRuntime6(){
        boolean result;
        try{
            result = jrePackage.isCompatibleWith("1.6");
        }catch(NumberFormatException e){
            assert false;
            return false;
        }
        return result;
    }

    /**
     * VMごとプログラムを終了する。
     * ※おそらく随所でシャットダウンフックが起動されるはず。
     * 
     * @param exitCode 終了コード
     * @throws java.lang.SecurityException
     */
    public static void exit(int exitCode) throws SecurityException{
        try{
            runtime.exit(exitCode);
        }catch(SecurityException e){
            logger.log(Level.WARNING, "VMを終了できません。", e);
            throw e;
        }
        return;
    }

    /**
     * Jindolf の実行が可能な環境でなければ、即時VM実行を終了する。
     */
    private static void testEnvironment(){
        if(!hasRuntime5()){
            logger.severe(
                    Jindolf.title
                    + " needs JRE1.5 platform or later.");
            exit(1);
        }

        if(GraphicsEnvironment.isHeadless()){
            logger.severe(
                    Jindolf.title
                    + " needs bit-map display, keyboard, & pointing-device.");
            exit(1);
        }

        if(   '狼' != 0x72fc
           || '　' != 0x3000
           || '~'  != 0x007e
           || '\\' != 0x005c
           || '～' != 0xff5e ){
            logger.severe(
                    "ソースコードの文字コードが正しくコンパイルされていないかも。"
                   +"あなたは今、オリジナル開発元の意図しない文字コード環境で"
                   +"コンパイルされたプログラムを起動しようとしているよ。"
                   +"ソースコードの入手に際して"
                   +"どのような文字コード変換が行われたか認識している？"
                   +"コンパイルオプションで正しい文字コードを指定したかな？"
                    );
            exit(1);
        }
        
        return;
    }

    /**
     * Jindolf のスタートアップエントリ
     *  
     * @param args コマンドライン引数
     */
    public static void main(final String[] args){
        runtime.addShutdownHook(
            new Thread(){
                @Override
                public void run(){
                    System.out.flush();
                    System.err.flush();
                    runtime.runFinalization();
                    Thread.yield();
                    runtime.gc();
                    Thread.yield();
                    return;
                }
            }
        );
                
        testEnvironment();
        
        // Metal L&F でもの凄く日本語が汚くなるのを強制回避
        UIManager.put("swing.boldMetal", Boolean.FALSE);

        Window splashWindow = null;
        if(!hasRuntime6()){
            splashWindow = createSplashWindow();
            splashWindow.setVisible(true);
        }

        GUIUtils.dispatchAWTEvent();
        
        try{
            SwingUtilities.invokeAndWait(
                new Runnable(){
                    public void run(){
                        startGUI(args);
                        return;
                    }
                }
            );
        }catch(Exception e){
            logger.log(Level.SEVERE,
                       "アプリケーション初期化に失敗しました",
                       e);
            exit(1);
        }

        if(splashWindow != null){
            splashWindow.setVisible(false);
            splashWindow.dispose();
            splashWindow = null;
        }
        
        GUIUtils.dispatchAWTEvent();

        return;
    }

    /**
     * AWTイベントディスパッチスレッド版スタートアップエントリ
     * 
     * @param args コマンドライン引数
     */
    // TODO コマンドライン解析
    private static void startGUI(String[] args){
        preLoadClass();

        LandsModel model = new LandsModel();
        try{
            model.loadLandList();
        }catch(IOException e){
            logger.log(Level.SEVERE,
                       "国定義一覧の読み込みに失敗しました",
                       e);
            exit(1);
        }
        
        TopFrameView topFrame = buildMVC(model);

        topFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        topFrame.setResizable(true);
        topFrame.setIconImage(GUIUtils.getWindowIconImage());
        topFrame.pack();
        topFrame.setSize(initGeometry);
        topFrame.setLocationByPlatform(true);
        Toolkit kit = topFrame.getToolkit();
        kit.setDynamicLayout(false);

        topFrame.setVisible(true);

        return;
    }

    /**
     * スプラッシュウィンドウを作成する。
     * 
     * @return 未表示のスプラッシュウィンドウ。
     */
    private static Window createSplashWindow(){
        JWindow splashWindow = new JWindow();
        JLabel splashLabel = new JLabel(GUIUtils.getLogoIcon());
        splashWindow.add(splashLabel);

        splashWindow.pack();
        splashWindow.setLocationRelativeTo(null); // locate center

        return splashWindow;
    }

    /**
     * モデル・ビュー・コントローラの結合?
     * 
     * @param model 最上位のデータモデル
     * @return 最上位ビュー画面
     */
    private static TopFrameView buildMVC(LandsModel model){
        MenuManager menuManager = new MenuManager();
        TopFrameView topFrame = new TopFrameView();
        topFrame.setJMenuBar(menuManager.createMenuBar());
        new Controller(menuManager, topFrame, model);
        return topFrame;
    }

    /**
     * hidden default constructor
     */
    private Jindolf(){
        super();
        return;
    }
}
